# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from decimal import Decimal
from typing import List
from typing import Union

from obm import connectors
from obm import utils


class SyncContextManagerMixin:
    def __enter__(self):
        return utils.sync_run(self.__aenter__())

    def __exit__(self, exc_type, exc, tb):
        return utils.sync_run(self.__aexit__(exc_type, exc, tb))


class NodeMixin(SyncContextManagerMixin):

    __connector = None

    @property
    def connector(self):
        if self.__connector:
            return self.__connector
        self.__connector = connectors.MAPPING[self.name](
            self.rpc_host,
            self.rpc_port,
            self.rpc_username,
            self.rpc_password,
            self.loop,
            self.session,
            self.timeout,
        )
        return self.__connector

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def open(self):
        await self.connector.open()

    async def close(self):
        await self.connector.close()

    async def get_latest_block_number(self):
        return await self.connector.latest_block_number

    async def create_address(self, password: str = "") -> str:
        return await self.connector.create_address(password)

    async def estimate_fee(
        self,
        from_address: str = None,
        to_address: str = None,
        amount: str = None,
        fee: Union[dict, Decimal] = None,
        data: str = None,
        conf_target: int = 1,
    ) -> Decimal:
        return await self.connector.estimate_fee(
            from_address, to_address, amount, fee, data, conf_target,
        )

    async def fetch_recent_transactions(
        self, limit: int = 10, **kwargs,
    ) -> List[dict]:
        return await self.connector.fetch_recent_transactions(limit, **kwargs)

    async def send_transaction(
        self,
        amount: Union[Decimal, float],
        to_address: str,
        from_address: str = None,
        fee: Union[dict, Decimal] = None,
        password: str = "",
        subtract_fee_from_amount: bool = False,
    ) -> dict:
        return await self.connector.send_transaction(
            amount,
            to_address,
            from_address,
            fee,
            password,
            subtract_fee_from_amount,
        )

    async def fetch_in_wallet_transaction(
        self, txid: str
    ) -> dict:
        """Fetches the transaction by txid from a blockchain.

        Args:
            txid: Transaction ID to return.

        Returns:
            Dict that represent the transaction.
        """
        return await self.connector.fetch_in_wallet_transaction(txid)

    async def fetch_in_wallet_transactions(
        self, txids: List[str]
    ) -> List[dict]:
        """Fetches the transactions by txids from a blockchain.

        Args:
            txids: Transaction IDs to return.

        Returns:
            Dict that represent the transactions list.
        """
        return await self.connector.fetch_in_wallet_transactions(txids)


class TransactionMixin:
    async def sync(self):
        """Synchronizes the transaction with blockchain.

        Make sense to synchronize only a transaction that still have
        block_number equals to None. Because a transaction that have already
        added to block is unchanging by blockchain design.

        Returns:
            Synchronized transaction.
        """
        if self.block_number is None:
            tx = await self.node.fetch_in_wallet_transaction(txid=self.txid)
            self.block_number = tx["block_number"]
        return self



# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
from decimal import Decimal
from typing import Optional, Union

import aiohttp

from obm import connectors, mixins, validators

__all__ = [
    "Currency",
    "Node",
]


class Currency:
    def __init__(self, name: str):
        if not isinstance(name, str):
            raise TypeError(
                f"Name must be a string, not '{type(name).__name__}'"
            )
        self.name = validators.validate_currency_is_supported(name)

    @classmethod
    def create_for(cls, connector_name: str):
        return cls(name=connectors.MAPPING[connector_name].currency)


class Node(mixins.NodeMixin):
    def __init__(
        self,
        name: str,
        currency: Currency = None,
        rpc_host: str = "localhost",
        rpc_port: Optional[int] = None,
        rpc_username: Optional[str] = None,
        rpc_password: Optional[str] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        session: Optional[aiohttp.ClientSession] = None,
        timeout: Union[int, float] = connectors.DEFAULT_TIMEOUT,
    ):
        if not isinstance(name, str):
            raise TypeError(
                f"Name must be a string, not '{type(name).__name__}'"
            )
        self.name = validators.validate_node_is_supported(name)
        self.currency = currency or Currency.create_for(name)
        self.rpc_port = rpc_port
        self.rpc_host = rpc_host
        self.rpc_username = rpc_username
        self.rpc_password = rpc_password
        self.loop = loop
        self.session = session
        self.timeout = timeout
        # This statement is necessary to perform validation
        assert self.connector.node == self.name
        super().__init__()


class Transaction(mixins.TransactionMixin):
    def __init__(
        self,
        node: Node,
        to_address: str,
        amount: Union[float, Decimal],
        from_address: Optional[str] = None,
        block_number: Optional[int] = None,
        txid: Optional[str] = None,
        category: Optional[str] = None,
        timestamp: Optional[int] = None,
        fee: Optional[Union[dict, float, Decimal]] = None,
    ):
        # TODO: Add validation
        self.node = node
        self.to_address = to_address
        self.amount = amount
        self.from_address = from_address
        self.block_number = block_number
        self.txid = txid
        self.category = category
        self.timestamp = timestamp
        self.fee = fee
        super().__init__()



# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
class BaseError(Exception):
    """Base connector error."""


class NodeError(BaseError):
    """Node returns an error."""


class NodeInvalidResponceError(NodeError):
    """Node returns a invalid response."""


class NodeUnsupportedError(NodeError, ValueError):
    """Node unsupported error."""

class NodeTooSmallTransactionAmount(NodeError):
    """Insufficient transaction amount for substract fee. """


class CurrencyUnsupportedError(NodeError, ValueError):
    """Currency unsupported error."""


class NetworkError(BaseError):
    """Node network error."""


class NetworkTimeoutError(NetworkError):
    """Node request riches timeout."""



# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import asyncio
import functools
import json
from decimal import Decimal
from typing import List, Optional, Union

import aiohttp

from obm import exceptions

DEFAULT_TIMEOUT = 5 * 60


def _catch_network_errors(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except aiohttp.ServerTimeoutError:
            self = args[0]
            raise exceptions.NetworkTimeoutError(
                f"The request to node was longer "
                f"than timeout: {self.timeout}"
            )
        except aiohttp.ClientError as exc:
            raise exceptions.NetworkError(exc)

    return wrapper


class _DecimalEncoder(json.JSONEncoder):
    def default(self, obj):  # pylint: disable=method-hidden, arguments-differ
        if isinstance(obj, Decimal):
            return str(obj)
        return super().default(obj)


class Connector(abc.ABC):
    def __init__(
        self,
        rpc_host: str,
        rpc_port: int,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        session: Optional[aiohttp.ClientSession] = None,
        timeout: Union[int, float] = DEFAULT_TIMEOUT,
    ):
        if not isinstance(rpc_host, str):
            raise TypeError(
                f"PRC host must be a string, not '{type(rpc_host).__name__}'"
            )
        if not isinstance(rpc_port, int):
            raise TypeError(
                f"PRC port must be an integer, not '{type(rpc_port).__name__}'"
            )
        if session is not None:
            if not isinstance(session, aiohttp.ClientTimeout):
                raise TypeError(
                    f"Session must be a aiohttp.ClientSession, "
                    f"not '{type(session).__name__}'"
                )
        if timeout is not None:
            if not isinstance(timeout, (float, int)):
                raise TypeError(
                    f"Timeout must be a number, not '{type(timeout).__name__}'"
                )
            if timeout <= 0:
                raise ValueError("Timeout must be greater than zero")

        # TODO: Create auth here
        url = f"{rpc_host}:{rpc_port}"
        self.rpc_host = rpc_host
        self.rpc_port = rpc_port
        self.timeout = timeout
        self.session = session
        self.url = url if url.startswith("http") else "http://" + url
        self.loop = loop or asyncio.get_event_loop()

    def __getattribute__(self, item):
        if item != "METHODS" and item in self.METHODS:
            return functools.partial(self.wrapper, method=self.METHODS[item])
        return super().__getattribute__(item)

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    async def open(self):
        if self.session is None:
            self.session = aiohttp.ClientSession(
                loop=self.loop,
                headers=self.headers,
                auth=self.auth,
                timeout=aiohttp.ClientTimeout(total=self.timeout),
                json_serialize=functools.partial(
                    json.dumps, cls=_DecimalEncoder
                ),
            )

    async def close(self):
        if self.session is not None:
            await self.session.close()
            self.session = None

    @_catch_network_errors
    async def call(self, payload: dict) -> dict:
        await self.open()
        async with self.session.post(  # type: ignore
            url=self.url,
            json=payload,
            raise_for_status=True,
        ) as response:
            return await response.json(
                loads=functools.partial(json.loads, parse_float=Decimal)
            )

    @staticmethod
    async def validate(response: dict) -> Union[dict, list]:
        try:
            if error := response.get("error"):
                raise exceptions.NodeError(error)
            return response["result"]
        except KeyError:
            raise exceptions.NodeInvalidResponceError(response)

    @property
    @abc.abstractmethod
    def node(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def currency(self) -> str:
        ...

    @abc.abstractmethod
    async def wrapper(self, *args, method: str = None) -> Union[dict, list]:
        ...

    # Unified interface

    @property
    @abc.abstractmethod
    async def latest_block_number(self) -> int:
        ...

    @abc.abstractmethod
    async def create_address(self, password: str = "") -> str:
        return await self.rpc_personal_new_account(password)

    @abc.abstractmethod
    async def estimate_fee(
        self,
        from_address: str = None,
        to_address: str = None,
        amount: str = None,
        fee: Union[dict, Decimal] = None,
        data: str = None,
        conf_target: int = 1,
    ) -> Decimal:
        ...

    @abc.abstractmethod
    async def send_transaction(
        self,
        amount: Union[Decimal, float],
        to_address: str,
        from_address: str = None,
        fee: Union[dict, Decimal] = None,
        password: str = "",
        subtract_fee_from_amount: bool = False,
    ) -> dict:
        ...

    @abc.abstractmethod
    async def fetch_recent_transactions(
        self, limit: int = 10, **kwargs,
    ) -> List[dict]:
        """Fetches most recent transactions from a blockchain.

        Args:
            limit: The number of transactions to return. Defaults to 10.

        Returns:
            Most recent transactions list.
        """

    @abc.abstractmethod
    async def fetch_in_wallet_transaction(self, txid: str) -> dict:
        """Fetches the transaction by txid from a blockchain.

        Args:
            txid: Transaction ID to return.

        Returns:
            Dict that represent the transaction.
        """

    async def fetch_in_wallet_transactions(
        self, txids: List[str]
    ) -> List[dict]:
        """Fetches the transactions by txids from a blockchain.

        Args:
            txids: Transaction IDs to return.

        Returns:
            Dict that represent the transactions list.
        """
        batch = [self.fetch_in_wallet_transaction(txid) for txid in txids]
        return await asyncio.gather(*batch)



import asyncio
import inspect


def _get_or_create_event_loop():
    try:
        asyncio.get_running_loop()
        return asyncio.get_event_loop()
    except RuntimeError:
        # The OS thread is not main
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop


def sync_run(coro, loop=None):
    assert inspect.iscoroutine(coro)
    loop = loop or _get_or_create_event_loop()
    if loop.is_running():
        return coro
    return loop.run_until_complete(coro)



# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from obm.connectors.base import DEFAULT_TIMEOUT, Connector
from obm.connectors.bitcoin import BitcoinCoreConnector
from obm.connectors.ethereum import GethConnector

__all__ = [
    "BitcoinCoreConnector",
    "GethConnector",
    "DEFAULT_TIMEOUT",
]

MAPPING = {
    "bitcoin-core": BitcoinCoreConnector,
    "geth": GethConnector,
}
SUPPORTED_CURRENCIES = list({node.currency for node in MAPPING.values()})



# Copyright 2020 Alexander Polishchuk
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Type

from obm import connectors, exceptions


def validate_node_is_supported(
    name: str, error_cls: Type[Exception] = exceptions.NodeUnsupportedError
) -> str:
    supported_nodes = list(connectors.MAPPING)
    if name not in supported_nodes:
        raise error_cls(f"Unsupported node. Available only: {supported_nodes}")
    return name


def validate_currency_is_supported(
    name: str, error_cls: Type[Exception] = exceptions.CurrencyUnsupportedError
) -> str:
    if name not in connectors.SUPPORTED_CURRENCIES:
        raise error_cls(
            f"Unsupported currency. Available only: "
            f"{connectors.SUPPORTED_CURRENCIES}"
        )
    return name
